{{
*************************************
* bootconf module
*************************************

{{

How this code is designed to work:

1. Bootconf uses serial port, eeprom, and SDIO card by default. It lets users create
a configuration for devices that is saved on eeprom. User can test SDIO access with
the 'D' command, but booting a binary is done with the bootloader program.

2. Bootconf keeps configuration pins for SDIO card in EEPROM configuration space $C0-$1AF.
Configuration space definition can be shared with applications with bootdef.spin. 

3. The first time the code boots or any time the EEPROM is reprogrammed, configuration
settings are displayed, and the user is allowed to edit the EEPROM data.

}}


OBJ
  pd    : "PinDefs"
  bd    : "bootdef"
  ctype : "ctype"
  sconv : "itoatoi"
  sf    : "snprintf"
  crc   : "CRC16_Slow"
  eep   : "EEPROM"
  fds   : "FullDuplexSerial"
  bsd   : "fsrw"

CON ' PinDefs
   _clkmode  = pd#_clkmode
   _xinfreq  = pd#_xinfreq

  BS = 8
  LF = $A
  CR = $D
  COGMAX = 8
  
VAR
  byte cbuf[33]       ' for itoatoi

DAT
  DefaultProduct_str    byte "Propeller", 0
  TV_BASEPIN_CFG_STR    byte "TV  Basepin [#/n]", 0
  VGA_BASEPIN_CFG_STR   byte "VGA Basepin [#/n]", 0
  NTSCPAL_CFG_STR       byte "NTSC? [Y/n]", 0
  SDCARD_BASEPIN_CFG_STR byte "SD Card Basepin [#/n]", 0
  MOUSE_DAT_PIN_CFG_STR byte "Mouse DAT [#/n]", 0
  MOUSE_CLK_PIN_CFG_STR byte "Mouse CLK [#/n]", 0
  KEYBD_DAT_PIN_CFG_STR byte "Keybd DAT [#/n]", 0
  KEYBD_CLK_PIN_CFG_STR byte "Keybd CLK [#/n]", 0
  ENET_CS_PIN_CFG_STR   byte "ENET CS  [#/n]", 0
  ENET_SCK_PIN_CFG_STR  byte "ENET SCK [#/n]", 0
  ENET_SI_PIN_CFG_STR   byte "ENET SI  [#/n]", 0
  ENET_SO_PIN_CFG_STR   byte "ENET SO  [#/n]", 0

CON
  CFG_RDONLY = 0

VAR ' Config Version 1_0001 overlay ....

  long cfg_size                ' Part of bootloader header ... keep as size-4 to exclude crc16      
  long cfg_version             ' Part of bootloader header ... Version number represented as $HHdddd H is header version, d is data version
  long cfg_sdpins              ' Part of bootloader header
  long cfg_baud_rate           ' Part of bootloader header
  long cfg_sdboot_fname[4]     ' Part of bootloader header ... 16 chars
  long cfg_pingrp1[2]          ' offset $10: tv base, vga base, ntsc/pal, sdcard base, mouse dat, mouse clk, kb dat, kb clk
  long cfg_pingrp2[2]          ' ethernet pins: i_cs, i_sck, i_si, i_so, i_int
  long cfg_pingrpRSV[12]        '  
  long cfg_prodname[4]         ' 32 chars ... product name or copyright  
  long cfg_userid[4]           ' 16 chars 
  long cfg_passwd[4]           ' 16 chars 
  long cfg_hostid[4]           ' 16 chars 
  long cfg_ipv4_address        '   
  long cfg_ipv4_netmask        ' 
  long cfg_ipv4_gateway        ' 
  long cfg_ipv4_dns            ' 
  long cfg_macaddr[2]          ' 8 bytes ... only need 6 
  long cfg_rsv2_data[2]        ' 8 bytes not defined
  long cfg_bootp_fname[8]      ' 32 bytes     
  long cfg_rsv3[3]             ' 12 bytes not defined
  long cfg_crc16               '

  long cfg_len
  long fdscog                   ' don't print if this is 0
    
PUB main | rc

  eep.start(pd#EEPROM1_SCL, pd#EEPROM1_DAT)
  rc := getEEconfig(@cfg_size, @cfg_crc16-@cfg_size+4)

  if rc ' badcrc
    rc := fds.start(pd#SERIAL1_RX, pd#SERIAL1_TX, pd#SERIAL1_MODE, pd#SERIAL1_BAUD)
  else
    rc := fds.start(pd#SERIAL1_RX, pd#SERIAL1_TX, pd#SERIAL1_MODE, cfg_baud_rate)

  fdscog := rc
  
  waitcnt(clkfreq+cnt)

  printf(string(CR,"Serial port started on COG %d"),parm(rc))

  cfg_len := @cfg_crc16-@cfg_size
  
  biosConfig(true)
  

PRI dba(addr)
  return addr
  
PRI dir | sdpin, sbuf[100]

  result := sdopen
  if result > -1
    bsd.opendir
    putchar(CR)
    repeat while bsd.nextfile(@sbuf) == 0
      printf(string(" %s\n"),parm(@sbuf))

VAR long cogTestStack[4]

PRI cogTest

PRI getFreeCog
  result := cognew(cogtest,@cogTestStack)
  waitcnt(clkfreq/10+cnt)
  cogstop(result)


PRI newEEconfig | ii, len, adj

  puts(string(CR,"Preparing new config ..."))
  waitcnt(clkfreq+cnt)

  bytefill(dba(@cfg_size), 0, cfg_len+4)

  long[dba(@cfg_size)]         := cfg_len
  long[dba(@cfg_version)]      := $0000     ' 0 for now, $101 after published
  bytemove(dba(@cfg_prodname), @DefaultProduct_str, strsize(@DefaultProduct_str)+1) ' copy null too
  
  byte[dba(@cfg_sdpins)  + bd#SDCARD_BASEPIN_BYTE]              := pd#FSRW1_DO

  byte[dba(@cfg_pingrp1) + bd#TV_BASEPIN_BYTE]                  := pd#TV_DAC1
  byte[dba(@cfg_pingrp1) + bd#VGA_BASEPIN_BYTE]                 := pd#VGA1
  byte[dba(@cfg_pingrp1) + bd#NTSCPAL_BYTE]                     := pd#NTSC
  byte[dba(@cfg_pingrp1) + bd#MOUSE_DAT_PIN_BYTE]               := pd#MOUSE1_DAT
  byte[dba(@cfg_pingrp1) + bd#MOUSE_CLK_PIN_BYTE]               := pd#MOUSE1_CLK
  byte[dba(@cfg_pingrp1) + bd#KEYBD_DAT_PIN_BYTE]               := pd#KEYBOARD1_DAT
  byte[dba(@cfg_pingrp1) + bd#KEYBD_CLK_PIN_BYTE]               := pd#KEYBOARD1_CLK

  byte[dba(@cfg_pingrp2) + bd#ENET_CS_PIN_BYTE]                 := pd#ENC1_CS
  byte[dba(@cfg_pingrp2) + bd#ENET_SCK_PIN_BYTE]                := pd#ENC1_SCK
  byte[dba(@cfg_pingrp2) + bd#ENET_SI_PIN_BYTE]                 := pd#ENC1_SI
  byte[dba(@cfg_pingrp2) + bd#ENET_SO_PIN_BYTE]                 := pd#ENC1_SO
  
  long[dba(@cfg_crc16)] := crc.get(dba(@cfg_size), @cfg_crc16-@cfg_size, crc#CRC_INIT)

   
PRI getEEconfig(dptr,length) | mycrc, ii, rc
' read database

  printf(string(CR,"Load EEPROM config $%04x - %d bytes"),parm(dptr)+parm(length))

  eep.Initialize
  repeat ii from 0 to length-1 step eep#PageSize
    rc := eep.ReadPage(eep#EEPROM, bd#BIOS_DBASE_ADDR+ii, dptr+ii, eep#PageSize)
    if rc
      printf(string(CR,"Read Error $d. Addr $%x\n"),parm(rc)+parm(bd#BIOS_DBASE_ADDR+ii))
      return rc
      
  'dumpbytes(dptr, length)                                   
    
  ' start with -1 to detect blank array
  mycrc := crc.get(dptr, @cfg_crc16-@cfg_size, crc#CRC_INIT)

  if(long[dptr+length-1] <> mycrc)
    printf(string(CR,"Bad CRC $%04x. Expected $%04x\n"),parm(long[dptr+length-1])+parm(mycrc))

  'dumpEEconfigBytes(dptr, length)
  
  if(long[dptr+length-1] <> mycrc)
    return -1
  return 0    

PRI saveEEconfig(dptr,length) | rc, ii
' write database

  printf(string(CR,"Saving EEPROM config - %d bytes"),parm(length))

  eep.Initialize
  repeat ii from 0 to length-1 step eep#PageSize
    rc := eep.WritePage(eep#EEPROM, bd#BIOS_DBASE_ADDR+ii, dptr+ii, eep#PageSize)
    repeat while eep.WriteWait(eep#EEPROM, bd#BIOS_DBASE_ADDR+ii)
    if rc
      printf(string(CR,"Write Error $d. Addr $%x\n"),parm(rc)+parm(dba(@cfg_size)+ii))
      return rc
    
  'getEEconfig(dptr,length)
  
  return 0    

PRI dumpEEconfigBytes(dptr, length) | ii
  'tv.out(CR)
  repeat ii from 0 to length-1 step 4
    ifnot ii // 16
      'printf(string("\n%04x "), parm(dba(@cfg_size)+ii))
      printf(string("\n%04x "), parm(dptr+ii))
    printf(string("%08x"), parm(long[dptr+ii]))
    ifnot ii // 16 == 15
      putchar(" ")
    'if (ii // 192) == 188
    '  repeat
  putchar($d)  


PRI dumpBytes(dptr, length) | ii
  repeat ii from 0 to length-1 step 1
    ifnot ii // 16
      putchar(CR)
      fds.hex(dptr+ii,4)
      putchar(" ")
    fds.hex(byte[dptr+ii],2)
    'putchar(" ")
    ifnot ii // 16 == 15
      putchar(" ")
  putchar(CR)
  
PRI biosConfig(force) | len, input, addr, data, str, ptr, sbuf[100], unk

  input := 0
  if force <> true
    repeat 2
      waitcnt(clkfreq+cnt)
      input := fds.rxcheck
      if input == 3 or input == $1B
        quit

  if getEEconfig(dba(@cfg_size), @cfg_crc16-@cfg_size+4) <> 0
    newEEconfig

  ' if config is blank, load defaults
  if long[dba(@cfg_crc16)] == 0
    newEEconfig
  elseif force <> true
    if input <= 0
      'print(string(CR,"Attempting normal boot ...",CR))
      return
        
  dumpEEconfigInfo(0)

  addr := dba(@cfg_userid)
  data := 0
  str  := 0

  repeat

    print(string(CR,"bootconf> "))
        
    input := getChar
    if input <> CR
      putchar(input)
       
    case input

      "a":
        addr := getAddr(@cbuf) 

      "c":
        printf(string(CR,"Last free COG %d",CR),parm(getFreeCog))

      "d": str  := 0
        data := getData(@cbuf,16) 

      "s":
        str  := getString(@sbuf,100,CR)

      "r":
        showField(addr)

      "w":
        setField(addr,str,data)

      "D":
        dir

      "C":
        dumpEEconfigInfo(true) ' use guided config

      "S":
        dumpEEconfigBytes(dba(@cfg_size), @cfg_crc16-@cfg_size+4)
        dumpEEconfigInfo(false) ' just show settings

      "E":
        len := (@cfg_crc16-@cfg_size+4)
        printf(string("Erasing config space %d\n"),parm(len))
        bytefill(dba(@cfg_size), 0, len)
        dumpEEconfigBytes(dba(@cfg_size), len)
        saveEEconfig(dba(@cfg_size), @cfg_crc16-@cfg_size+4)

      "L":
        getEEconfig(dba(@cfg_size), @cfg_crc16-@cfg_size+4)

      "N":
        newEEconfig

      "W":
        long[dba(@cfg_crc16)] := crc.get(dba(@cfg_size), @cfg_crc16-@cfg_size, crc#CRC_INIT)
        printf(string("\nWriting config with CRC $%04x"),parm(long[dba(@cfg_crc16)]))
        saveEEconfig(dba(@cfg_size), @cfg_crc16-@cfg_size+4)

      "q":
        print(string(CR,"Save [y/n] ? "))
        if getchar == "y"
          saveEEconfig(dba(@cfg_size), @cfg_crc16-@cfg_size+4)
        reboot  
        quit

      "?":
        print(string(CR,"a: set address for data read/write"))
        print(string(CR,"s: set string data, d: set number"))
        print(string(CR,"r: read one RAM address"))
        print(string(CR,"w: write one RAM address"))
        print(string(CR,"D: dump SD directory"))
        print(string(CR,"C: start config utility"))
        print(string(CR,"S: show RAM data and config fields"))
        print(string(CR,"E: erase EEPROM space, L: load EEPROM"))
        print(string(CR,"N: make new config in RAM"))
        print(string(CR,"W: write RAM config to EEPROM"))
        print(string(CR,"?: show help and addr/data settings"))
        print(string(CR,"q: reboot after optionally writing settings"))

        puts(string(CR,CR,"Current r/w settings:"))
        printf(string("  addr $%04x, "),parm(addr))
        if strsize(str)
          printf(string("str  '%s'\n"),parm(str))
        else
          printf(string("data $%08x\n"),parm(data))


PRI getAddr(buf) | ch, ii
  ii := 0
  repeat 8
    ch := getchar
    if ch == CR
      quit
    if ch == BS and ii > 0
      ii--
    else
      byte[buf+ii] := ch
      ii++
    byte[buf+ii] := 0
    putchar(ch)
  return sconv.atoib(buf,16)
  
PRI getData(buf,base) | ch, ii
  ii := 0
  repeat 8
    ch := getchar
    if ch == CR
      quit
    if ch == BS and ii > 0
      ii--
    else
      byte[buf+ii] := ch
      ii++
    byte[buf+ii] := 0
    putchar(ch)
  return sconv.atoib(buf,base)

PRI getString(buf,lim,trm) | ch, ii
  ii := 0
  byte[buf] := 0
  repeat
    ch := getchar
    if ch == trm
      quit
    if ch == BS and ii > 0
      ii--
    else
      byte[buf+ii] := ch
      ii++
    byte[buf+ii] := 0
    putchar(ch)
    if ii => lim
      quit
  return buf

PRI showField(addr)
  'printf(string("p %04x strlen(p) %d *p %08x\n"),parm(addr)+parm(strsize(addr))+parm(long[addr]))
  if addr
    if strsize(addr) and ctype.isalnum(byte[addr])
      print(addr)
    else
      printf(string("%08x"),parm(long[addr]))
   
PRI setField(addr,str,data) | ii 
  if addr
    if str
      ii := 0
      repeat
        byte[addr+ii] := byte[str+ii]
        ii++
      while(byte[str+ii])
      byte[addr+ii] := 0
    else
      long[addr] := data


{
#ifdef USE_CONSOLE_CONFIG
}
PRI configStringField(set, field, len, pnull) | ii, s, sbuf[100]
  repeat ii from 0 to len-1
    if ctype.isalnum(byte[field+ii])
      putchar(byte[field+ii])
    if byte[field+ii] == 0 and pnull == 0
      quit
  if(set)
    putchar(">")
    s := getString(@sbuf,len,CR)
    if strsize(s)
      setField(field,s,0)

PRI configNumberField(set, field, base) | s
  print(sconv.itoab(long[field], @cbuf, base))
  if(set)
    putchar(">")
    s := getData(@cbuf,base)
    if strsize(@cbuf)
      setField(field,0,s)

PRI configField(set, name, field, len, pnull)
  printf(string(CR,"%s:%d:%04x: "),parm(name)+parm(len)+parm(field+bd#BIOS_DBASE_ADDR-dba(@cfg_size)))
  if len == 4
    configNumberField(set, field,16)
  else
    configStringField(set, field, len, pnull)
           
PRI configFieldDecimal(set, name, field, len, pnull)
  printf(string(CR,"%s:%d:%04x: "),parm(name)+parm(len)+parm(field+bd#BIOS_DBASE_ADDR-dba(@cfg_size)))
  if len == 4
    configNumberField(set, field, 10)
           
PRI configPinField(set, cname, field, pin, name) | ii, s, sbuf[10], adj, mypin
  printf(string(CR,"%s:1:%x:%s "),parm(cname)+parm(field+pin+bd#BIOS_DBASE_ADDR-dba(@cfg_size))+parm(name))
  mypin := byte[field+pin]

  if mypin > $40
    putchar(mypin)
  else
    print(sconv.itoa(mypin, @cbuf))

  if(set)
    putchar(">")
    s := getString(@sbuf,4,CR)
    'print(s)
    if strsize(@sbuf)
      case byte[s]
        "n","N":
          byte[field+pin] := "n"        
        "y","Y":
          byte[field+pin] := "Y"        
        other:
          if sconv.atoi(s) => 0
            byte[field+pin] := sconv.atoi(s) 

PRI configFieldPins(set, name, field, len, pnull) | adj, ii

  printf(string(CR,"%s:%d:%04x: "),parm(name)+parm(len)+parm(field+bd#BIOS_DBASE_ADDR-dba(@cfg_size)))

  if field == dba(@cfg_sdpins)
    configPinField(set, name, field, bd#SDCARD_BASEPIN_BYTE,     @SDCARD_BASEPIN_CFG_STR)

  if field == dba(@cfg_pingrp1)
    configPinField(set, name, field, bd#TV_BASEPIN_BYTE,         @TV_BASEPIN_CFG_STR)
    configPinField(set, name, field, bd#VGA_BASEPIN_BYTE,        @VGA_BASEPIN_CFG_STR)
    configPinField(set, name, field, bd#NTSCPAL_BYTE,            @NTSCPAL_CFG_STR)
    configPinField(set, name, field, bd#MOUSE_DAT_PIN_BYTE,      @MOUSE_DAT_PIN_CFG_STR)
    configPinField(set, name, field, bd#MOUSE_CLK_PIN_BYTE,      @MOUSE_CLK_PIN_CFG_STR)
    configPinField(set, name, field, bd#KEYBD_DAT_PIN_BYTE,      @KEYBD_DAT_PIN_CFG_STR)
    configPinField(set, name, field, bd#KEYBD_CLK_PIN_BYTE,      @KEYBD_CLK_PIN_CFG_STR)
  if field == dba(@cfg_pingrp2)                              
    configPinField(set, name, field, bd#ENET_CS_PIN_BYTE,        @ENET_CS_PIN_CFG_STR)
    configPinField(set, name, field, bd#ENET_SCK_PIN_BYTE,       @ENET_SCK_PIN_CFG_STR)
    configPinField(set, name, field, bd#ENET_SI_PIN_BYTE,        @ENET_SI_PIN_CFG_STR)
    configPinField(set, name, field, bd#ENET_SO_PIN_BYTE,        @ENET_SO_PIN_CFG_STR)


PRI firstIPv4byte(c) | s, fbuf[10], sbuf[10]
  putchar(c)
  fbuf[0] := c
  s := getString(@sbuf,4,".")
  bytemove(@fbuf+1,@sbuf,strsize(@sbuf))
  s := @fbuf
  putchar(".")
  return s

PRI configFieldIPv4(set, name, field, len, pnull) | adj, ii, s, c, fbuf[10], sbuf[10]
  printf(string(CR,"%s:%d:%04x: "),parm(name)+parm(len)+parm(field+bd#BIOS_DBASE_ADDR-dba(@cfg_size)))

  repeat ii from 0 to 4-1     ' IPv4 addrs are always 4 bytes
    print(sconv.itoab(byte[field+ii],@cbuf,10))
    if ii < 3
      putchar(".")
  if set
    putchar(">")
    'c := bser.rx
    c := getchar
    if c <> LF and c <> CR
      s := firstIPv4byte(c)
      if strsize(s)
        byte[field+0] := sconv.atoi(s)
      repeat ii from 1 to 4-1     ' IPv4 addrs are always 4 bytes
        if ii < 3
          s := getString(@sbuf,4,".")
          putchar(".")
        else
          s := getString(@sbuf,4,CR)
        if strsize(s)
          byte[field+ii] := sconv.atoi(s)
   
PRI configFieldMACADDR(set, name, field, len, pnull) | adj, ii, s, c, fbuf[10], sbuf[10]
  printf(string(CR,"%s:%d:%04x: "),parm(name)+parm(len)+parm(field+bd#BIOS_DBASE_ADDR-dba(@cfg_size)))

  repeat ii from 0 to 6-1     ' mac addrs are always 6 bytes
    printf(string("%02x"), parm(byte[field+ii]))
    if ii < 5
      putchar(".")
  if set
    putchar(">")
     'c := bser.rx
     c := getchar
    if c <> LF and c <> CR
      s := firstIPv4byte(c)
      if strsize(s)
        byte[field+0] := sconv.atoi(s)
      repeat ii from 1 to 6-1     ' mac addrs are always 6 bytes
        if ii < 5
          s := getString(@sbuf,4,".")
          putchar(".")
        else
          s := getString(@sbuf,4,CR)
        if strsize(s)
          byte[field+ii] := sconv.atoib(s,16)

    
PRI dumpEEconfigInfo(set)

  configField(CFG_RDONLY,string("cfg_size"),dba(@cfg_size),dba(@cfg_version)-dba(@cfg_size),0)
  configField(CFG_RDONLY,string("cfg_version"),dba(@cfg_version),dba(@cfg_sdpins)-dba(@cfg_version),0)
  configFieldPins(set,string("cfg_sdpins"),dba(@cfg_sdpins),dba(@cfg_baud_rate)-dba(@cfg_sdpins),0)
  configFieldDecimal(set,string("cfg_baud_rate"),dba(@cfg_baud_rate), dba(@cfg_sdboot_fname)-dba(@cfg_baud_rate), 0)
  configField(set,string("cfg_sdboot_fname"),dba(@cfg_sdboot_fname), dba(@cfg_pingrp1)-dba(@cfg_sdboot_fname),0)
  configFieldPins(set,string("cfg_pingrp1"),dba(@cfg_pingrp1), dba(@cfg_pingrp2)-dba(@cfg_pingrp1), 0)
  configFieldPins(set,string("cfg_pingrp2"),dba(@cfg_pingrp2), dba(@cfg_pingrpRSV)-dba(@cfg_pingrp2), 0)
  configField(set,string("cfg_prodname"),dba(@cfg_prodname), dba(@cfg_userid)-dba(@cfg_prodname), 0)
  configField(set,string("cfg_userid"),dba(@cfg_userid), dba(@cfg_passwd)-dba(@cfg_userid), 0)
  configField(set,string("cfg_passwd"),dba(@cfg_passwd), dba(@cfg_hostid)-dba(@cfg_passwd), 0)
  configField(set,string("cfg_hostid"),dba(@cfg_hostid), dba(@cfg_ipv4_address)-dba(@cfg_hostid), 0)
  configFieldIPv4(set,string("cfg_ipv4_address"),dba(@cfg_ipv4_address), dba(@cfg_ipv4_netmask)-dba(@cfg_ipv4_address), 0)
  configFieldIPv4(set,string("cfg_ipv4_netmask"),dba(@cfg_ipv4_netmask), dba(@cfg_ipv4_gateway)-dba(@cfg_ipv4_netmask),0)
  configFieldIPv4(set,string("cfg_ipv4_gateway"),dba(@cfg_ipv4_gateway), dba(@cfg_ipv4_dns)-dba(@cfg_ipv4_gateway),0)
  configFieldIPv4(set,string("cfg_ipv4_dns"),dba(@cfg_ipv4_dns), dba(@cfg_macaddr)-dba(@cfg_ipv4_dns),0)
  configFieldMACADDR(set,string("cfg_macaddr"),dba(@cfg_macaddr), dba(@cfg_bootp_fname)-dba(@cfg_macaddr),0)
  configField(set,string("cfg_bootp_fname"),dba(@cfg_bootp_fname), dba(@cfg_crc16)-dba(@cfg_bootp_fname),0)
  long[dba(@cfg_crc16)] := crc.get(dba(@cfg_size), @cfg_crc16-@cfg_size, crc#CRC_INIT)
  configField(CFG_RDONLY,string("cfg_crc16"),dba(@cfg_crc16), 4, 0)

{
#endif
//USE_CONSOLE_CONFIG
}

VAR byte cstrbuf[2]
PRI cstr(ch)
  byte[@cstrbuf+0] := ch
  byte[@cstrbuf+1] := 0
  return @cstrbuf
          
PRI putchar(ch)
  'printf(string("putchar %c\n"),parm(ch))
  if fdscog
    fds.str(cstr(ch))

PRI puts(str)
  if fdscog
    fds.str(str)
    fds.str(string(CR))

PRI print(str)
  if fdscog
    fds.str(str)
  
PRI printf(fmt, arg) | pb[256]
  sf.snprintf(@pb,255,fmt,arg)
  if fdscog
    fds.str(@pb)

PRI parm(v)
  return sf.p(v)
  

PRI getChar | id
  if fdscog
    return fds.rx

PRI sdopen | pin
  pin := byte[dba(@cfg_sdpins)+bd#SDCARD_BASEPIN_BYTE]
  result := \bsd.mount(pin)
  if result < 0
    printf(string("\nCan't mount SD Card: %d\n"), parm(pin))

 