{{
*************************************
* getconfig module convenience object
*************************************

{{

This convenience module is used to read config information from the EEPROM.
The Bootloader in EEPROM keeps a copy of this data in locations $C0-$1AF.

}}

{

' test code
'
CON ' PinDefs
   _clkmode  = pd#_clkmode
   _xinfreq  = pd#_xinfreq

CON
  BS = 8
  LF = $A
  CR = $D

OBJ
  fds : "FullDuplexSerial"
  sf  : "snprintf"
    
PUB main | badcrc, rc
  
  waitcnt(clkfreq+cnt)

  if init <> 0
    fds.start(pd#SERIAL1_RX, pd#SERIAL1_TX, pd#SERIAL1_MODE, pd#SERIAL1_BAUD) ' Start serial port with default baud
    waitcnt(clkfreq+cnt)
    printf(string(CR,"EEPROM Config Error.",CR),0)
    repeat

  rc := fds.start(pd#SERIAL1_RX, pd#SERIAL1_TX, pd#SERIAL1_MODE, getBaudRate)   ' Start serial port with config baud rate
  waitcnt(clkfreq+cnt)
  printf(string(CR,"Serial port started on COG %d"),p(rc))

  dumpBytes(@config, bd#BIOS_DBASE_SIZE)

  printf(string(" Config size %d\n"), p(getConfigSize))
  printf(string(" Config version major %d\n"), p(getVersionMajor))
  printf(string(" Config version minor %d\n"), p(getVersionMinor))
  printf(string(" SD base pin %d\n"), p(getSdBasePin))
  printf(string(" Console baud rate %d\n"), p(getBaudRate))
  printf(string(" SD boot file name %s\n"), p(getSdBootFilePtr))
  printf(string(" TV base pin %d\n"), p(getTvBasePin))
  printf(string(" VGA base pin %d\n"), p(getVgaBasePin))
  printf(string(" NTSC / PAL %c\n"), p(getNtscPal))
  printf(string(" Extra video byte %d\n"), p(getExtraVideo))
  printf(string(" Mouse dat pin %d\n"), p(getMouseDat))
  printf(string(" Mouse clk pin %d\n"), p(getMouseClk))
  printf(string(" Keybd dat pin %d\n"), p(getKeybdDat))
  printf(string(" Keybd clk pin %d\n"), p(getKeybdClk))
  printf(string(" Enet CS pin %d\n"), p(getEnetCS))
  printf(string(" Enet CK pin %d\n"), p(getEnetSCK))
  printf(string(" Enet SI pin %d\n"), p(getEnetSI))
  printf(string(" Enet SO pin %d\n"), p(getEnetSO))
  printf(string(" Product name %s\n"), p(getProdNamePtr))
  printf(string(" User ID %s\n"), p(getUserIdPtr))
  printf(string(" Password %s\n"), p(getPasswdPtr))
  printf(string(" Host ID %s\n"), p(getHostIdPtr))
  printf(string(" IPv4 address %x\n"), p(getIPv4_addr))
  printf(string(" IPv4 netmask %x\n"), p(getIPv4_netmask))
  printf(string(" IPv4 gateway %x\n"), p(getIPv4_gateway))
  printf(string(" IPv4 dnsaddr %x\n"), p(getIPv4_dns))
  printf(string(" MAC  address @%x\n"), p(getMacAddrPtr))
  printf(string(" Bootp file name %s\n"), p(getBootpFnamePtr))
  printf(string(" CRC16 %x\n"), p(getCRC16))

  repeat
  
PRI putchar(ch)
  fds.tx(ch)

PRI getChar | id
  return fds.rx

PRI print(str)
  fds.str(str)
  
PRI puts(str)
  print(str)
  putchar(CR)

PRI dumpBytes(dptr, length) | ii
  repeat ii from 0 to length-1 step 1
    ifnot ii // 16
      putchar(CR)
      fds.hex(dptr+ii,4)
      putchar(" ")
    fds.hex(byte[dptr+ii],2)
    'putchar(" ")
    ifnot ii // 16 == 15
      putchar(" ")
  putchar(CR)

PRI printf(fmt, arg) | pb[256]
  sf.snprintf(@pb,255,fmt,arg)
  print(@pb)

PRI p(v)
  return sf.p(v)

}

OBJ
  pd  : "PinDefs"
  bd  : "bootdef"
  crc : "CRC16_Slow"

VAR
  byte config[bd#BIOS_DBASE_SIZE+4]

PUB init
  'eep.start(pd#EEPROM1_SCL, pd#EEPROM1_DAT)
  return getEEconfig(@config, bd#BIOS_DBASE_SIZE)

PUB getEEconfig(dptr,length) | mycrc, ii, rc
' read database

  'printf(string(CR,"Load EEPROM config $%04x - %d bytes"),parm(dptr)+parm(length))
{
  eep.Initialize
  repeat ii from 0 to length-1 step eep#PageSize
    rc := eep.ReadPage(eep#EEPROM, bd#BIOS_DBASE_ADDR+ii, dptr+ii, eep#PageSize)
    if rc
      'printf(string(CR,"Read Error $d. Addr $%x\n"),parm(rc)+parm(bd#BIOS_DBASE_ADDR+ii))
      return rc
}
  ' By doing this we are requiring that the application be booted on top of a bootloader
  ' that automatically reads the config bytes into bd#BIOS_DBASE_ADDR otherwise it will fail.
  '
  repeat ii from 0 to length-1 step 4
    long[dptr+ii] := long[bd#BIOS_DBASE_ADDR+ii]
    
  'dumpbytes(dptr, length)                                   
    
  ' start with -1 to detect blank array
  mycrc := crc.get(dptr, bd#BIOS_DBASE_SIZE-4, crc#CRC_INIT)

  if(long[dptr+length-1] <> mycrc)
    'printf(string(CR,"Bad CRC $%04x. Expected $%04x\n"),parm(long[dptr+length-1])+parm(mycrc))

  'dumpEEconfigBytes(dptr, length)
  
  if(long[dptr+length-1] <> mycrc)
    return mycrc
    
  return 0    

PUB getConfigSize
'' Config size is a long that we can use for finding the CRC16. 
  return long[@config+bd#cfg_size]

PUB getVersionMajor
'' Version major is a word that describes the config data structure header version. 
  return word[@config+bd#cfg_version]

PUB getVersionMinor
'' Version minor is a word that describes the config data structure contents version. 
  return word[@config+bd#cfg_version+2]

PUB getSdBasePin
'' SD basepin is a byte that defines the start pin for the SD card interface.
'' Other SD pins can be defined in this series of 4 bytes if necessary.
  return byte[@config+bd#cfg_sdpins]

PUB getBaudRate
'' Baud rate is the rate that programs can use for console i/o.
  return long[@config+bd#cfg_baud_rate]

PUB getSdBootFilePtr
'' The SD boot file is the name of the file for the bootloader to run after power up or reset. 
  return @config+bd#cfg_sdboot_fname

PUB getTvBasePin
'' TV basepin is the first pin in a series to use for the TV DAC.
  return byte[@config+bd#cfg_pingrp1+bd#TV_BASEPIN_BYTE]

PUB getVgaBasePin
'' VGA basepin is the first pin in a series to use for the VGA interface.
  return byte[@config+bd#cfg_pingrp1+bd#VGA_BASEPIN_BYTE]

PUB getNtscPal
'' NTSC/PAL is a flag that allows selecting NTSC or PAL for the video standard.
  return byte[@config+bd#cfg_pingrp1+bd#NTSCPAL_BYTE]

PUB getExtraVideo
'' Extra video is just an extra byte in config data.
  return byte[@config+bd#cfg_pingrp1+bd#EXTRA_VIDEO_BYTE]

PUB getMouseDat
'' This defines the mouse data interface pin.
  return byte[@config+bd#cfg_pingrp1+bd#MOUSE_DAT_PIN_BYTE]

PUB getMouseClk
'' This defines the mouse clock interface pin.
  return byte[@config+bd#cfg_pingrp1+bd#MOUSE_CLK_PIN_BYTE]

PUB getKeybdDat
'' This defines the keyboard data interface pin
  return byte[@config+bd#cfg_pingrp1+bd#KEYBD_DAT_PIN_BYTE]

PUB getKeybdClk
'' This defines the keyboard clock interface pin.
  return byte[@config+bd#cfg_pingrp1+bd#KEYBD_CLK_PIN_BYTE]

PUB getEnetCS
'' This defines SDIO based Ethernet Chip Select interface pin.
  return byte[@config+bd#cfg_pingrp2+bd#ENET_CS_PIN_BYTE]

PUB getEnetSCK
'' This defines SDIO based Ethernet clock interface pin.
  return byte[@config+bd#cfg_pingrp2+bd#ENET_SCK_PIN_BYTE]

PUB getEnetSI
'' This defines SDIO based Ethernet serial in interface pin.
  return byte[@config+bd#cfg_pingrp2+bd#ENET_SI_PIN_BYTE]

PUB getEnetSO
'' This defines SDIO based Ethernet serial out interface pin.
  return byte[@config+bd#cfg_pingrp2+bd#ENET_SO_PIN_BYTE]

PUB getProdNamePtr
'' Product name is a user defined product string. 
  return @config+bd#cfg_prodname

PUB getUserIdPtr
'' User ID is for network or other login identification.
  return @config+bd#cfg_userid

PUB getPasswdPtr
'' Password is for network or other login identification.
  return @config+bd#cfg_passwd

PUB getHostIdPtr
'' Host ID is for network or other login identification.
  return @config+bd#cfg_hostid

PUB getIPv4_addr
'' IPv4 address is a 32 bit word that defines the device IP address.
  return long[@config+bd#cfg_ipv4_address]

PUB getIPv4_netmask
'' IPv4 netmask is a 32 bit word that defines the local IP address access range.
  return long[@config+bd#cfg_ipv4_netmask]

PUB getIPv4_gateway
'' IPv4 gateway is a 32 bit word that defines the IP address of an internet gateway server.
  return long[@config+bd#cfg_ipv4_gateway]

PUB getIPv4_dns
'' IPv4 dns is a 32 bit word that defines the IP address of the domain name server for host/address resolution.
  return long[@config+bd#cfg_ipv4_dns]

PUB getMacAddrPtr 
'' A MAC address is a 6 byte word; a pointer is returned so users can deal with it as they like.
  return @config+bd#cfg_macaddr

PUB getBootpFnamePtr
'' The bootp fname is the file name of a remote program to use for booting the device. This is a future feature. 
  return @config+bd#cfg_bootp_fname

PUB getCRC16
'' CRC16 is the data validator for the configuration space. An invalid CRC16 means config data is not usable. 
  return long[@config+bd#cfg_crc16]

   