{{
***********************************************
* sdxmodem.spin download to sdcard using xmodem
***********************************************
}}

{{

 Differences from Mike Cook's description below:

   PinDefs.spin are used for hardware and device definitions.
   Mike Green's code is not used since a spin boot is not required for just download.
   Tomas Rokicki's fsrw with faster sdcard asm access is used instead.

 =============================================================================

   File...... Xbee_Serial_XMODEM_Upload.spin
   Purpose... 
   Author.... Mike Cook
   E-mail.... mike-cook@tx.rr.com 
   Started... 03-14-2007
   Updated... 12-24-2007 
   Compiler.. Propeller Tool v1.05.8
   Platform.. Propeller Proto Board  Stock#: 32212    

 =============================================================================
   Credits:
   TopFile derived from:
                         Mike Green's FemtoBasic v2.002, with Tomas Rokicki's
                         FAT16 SD Support

                         Simon Ampleman's XModem Wave Player's XMODEM method.  



   Links..... SD Support:
              http://tomas.rokicki.com/fat16work/
              http://forums.parallax.com/forums/default.aspx?f=25&m=162235
              
              FemtoBasic:
              http://forums.parallax.com/forums/default.aspx?f=25&m=173975
              
              TV_Text:
              http://www.parallax.com/dl/sw/propeller/Setup-Propeller-Tool-v1.05.2.exe
              
              Propeller Tool:
              http://forums.parallax.com/forums/default.aspx?f=25&m=172361
              v1.05.4 - http://forums.parallax.com/forums/attach.aspx?a=12406
                                 
              XModem_Wave_Full:
              http://forums.parallax.com/forums/default.aspx?f=25&m=158620&p=1&ord=d
              

   Notes..... 1. To create a SPIN Binary:
                   a. Make sure your application's top file is selected in the
                      Propeller Tool's tabbed display
                   b. Press F8 in the Propeller Tool
                   c. Click on the "Save Binary File" button

              2. Prepare SD card:
                   a. Using your PC's OS; example: Windows XP; format a SD card
                      to have a FAT 16 file structure
                   b. Copy the file, created in step 1 item c, to the SD card
                   c. Rename the file to the DOS 8x3 file name format;
                      example: xxxxxxxx.xxx

              3. Load the Propeller platform:
                   a. Install the SD card in the Propeller platform 
                   b. Change the file name in this file to your file's name,
                      see DAT section, FileName
                   c. Press F11 to compile and load EEPROM


 -----[ History ]-------------------------------------------------------------

  12-27-2007 - Corrected problem in 256_FullDuplexSerial, this was causing
               the first XMODEM packet to be sent twice. Remove the 'flag' fix
               that was added to correct this issue.
               

  12-24-2007 - Could not figure out why both terminal programs were sending
               the first XMODEM packet twice. So shotgunned it to work, see
               notes below.

               About 4 hours of work, adding band aid, testing, deleting loads
               of debugging code, and then a full line by line comment.

  
  12-23-2007 - Modified SD Boot loader top file to include Simon Ampleman's
               XModem Wave Player's XMODEM method. Original program wrote data
               to an I2C EEPROM connected to PINs 0 & 1. Modified this routine
               to receive a file from a MaxStream Xbee connected to PINs 4 & 5
               and then store the file to a SD card using Mike Green's FemtoBasic
               v2.002, with Tomas Rokicki's FAT16 SD Support.

               About 8 hours of work, most of the time was spent trying to
               figure out why both windows terminal programs were failing the
               first XMODEM packet.
               

  03-14-2007 - Created this Top File to load a SPIN application from a SD card.
               Tested and added documentation.

  
 -----------------------------------------------------------------------------
 Program Description
 -----------------------------------------------------------------------------

  Testing Xbee XMODEM file transfer:

  Conditions: 1. Program starts and waits to receive a file via Xbee
        
              2. If file transfer times out or fails then goes to step 1
                    
              3. If successful, file is written to SD card and is
                 ASSUMED to be a *.SPIN file stored in *.binary format.
                 No checking of the file is done! If user uploads any
                 other file than a valid *.binary file, the boot loader
                 will try to load this file and run it.

              4. If the file is not evenly divisible by 128 bytes the
                 last packet will be padded with zero bytes. Does not seem
                 to cause a problem loading *.spin binary files in the
                 boot loader program.   

                       
 -----------------------------------------------------------------------------
}}


obj
   pd    : "PinDefs"
   sdfat : "fsrwqasm"                                      ' FAT File System Read/Write
   PC    : "FullDuplexSerial256"                           ' Need big serial buffer for xmodem file transfer
   cfg   : "GetConfig"
   
con
  _clkmode =     pd#_clkmode                               ' Set PLL to run at 16 times installed xtal (80MHz)
  _xinfreq =     pd#_xinfreq                               ' xtal frequency of 5MHz

  SOH      =     1                                         ' Packet Start
  EOT      =     4                                         ' End of Transmission
  ACK      =     6                                         ' Positive Acknowledgement
  NAK      =     21                                        ' Negative Acknowledgement
  CAN      =     24                                        ' Cancel
  PAD      =     26                                        ' Padding at the end of the file
      
  BS       =     8                                         ' ANSII backspace character
  CR       =     13                                        ' ANSII Carriage Return character
  LF       =     10                                        ' ANSII Line Feed Character
  CLS      =     12                                        ' ANSII Clear Screen character

  RX_Buffer=     20
  
       
var

  long ioControl[2]                                        ' Needed for fsrwFemto
  
  BYTE CRC                                                 ' XMODEM Protocol 
  BYTE PACKET                                              ' XMODEM Protocol 
  BYTE XRECV                                               ' XMODEM Protocol
  BYTE MODE                                                ' XMODEM Protocol
  BYTE RETRY                                               ' XMODEM Protocol
  BYTE XBUF[128]                                           ' XMODEM Protocol
  
  BYTE Rx_Data[RX_Buffer]
  

pub main | rc
  rc := cfg.init <> 0
  if rc
    PC.start(pd#SERIAL1_RX, pd#SERIAL1_TX, pd#SERIAL1_MODE, pd#SERIAL1_BAUD) ' Start serial port with default baud
    waitcnt(clkfreq+cnt)
    if \sdfat.mount(pd#FSRW1_DO) < 0                                  ' Attempt to mount SD card
      PC.str(@SdCardErr)                                     '  If SD mount failure then Display error
      repeat                                                  '  HALT PROGRAM
  else
    PC.start(pd#SERIAL1_RX, pd#SERIAL1_TX, pd#SERIAL1_MODE, cfg.getBaudRate)   ' Start serial port with config baud rate
    waitcnt(clkfreq+cnt)
    if \sdfat.mount(cfg.getSdBasePin) < 0                                  ' Attempt to mount SD card
      PC.str(@SdCardErr)                                     '  If SD mount failure then Display error
      repeat                                                  '  HALT PROGRAM
     
  repeat
    PC.str(string("SD Card Directory:",CR,CR))
    sdfat.opendir
    repeat while 0 == sdfat.nextfile(@RX_Data)
      PC.str(@RX_Data)
      PC.tx(CR)
    
    UPLOAD                                                 ' Go to UPLOAD method


PRI UPLOAD | TCNT , temp, ascii, len, ii

  PC.str(@Ready)                                           ' Tell sender we are ready to receive file

  PC.str(@GetName)

  if strsize(getString(@Rx_Data,20,CR)) == 0
    reboot
  
  PC.str(@IsAscii)

  ascii := 26       ' ascii eof $1a, ^Z
  if PC.rx <> "y"
    ascii := 0

  PC.str(string(CR,"Downloading "))
  if ascii
    PC.str(string("ASCII "))
  PC.str(@Rx_Data)
  PC.tx(CR)
  
  PACKET := 1                                              ' Transfer not yet started, Set Packet Number to first 
  RETRY  := 0                                              '                           Clear Retry count
  MODE   := 0                                              '                           Set Mode to zero
  
  REPEAT 80                                                ' 40 seconds max time to wait for sender to send the file
    PC.tx(NAK)                                             ' send NAK to sender to inform them we are ready to receive file
    XRECV := PC.rxtime (500)                               ' wait here for character or time-out

    IF (XRECV == SOH)                                      ' If character received is packet start
      MODE := 1                                            '   Set file transfer mode to state 1
      QUIT                                                 '   exit this repeat loop

  sdfat.popen(@RX_Data, "w")                               ' open file for write, if exists, then over write the file   
  
  REPEAT WHILE MODE == 1                                   ' While in file transfer state
    CRC := 0                                               ' clear CRC for later calculation 
    XRECV := PC.rx                                         ' get packet number

    IF (XRECV < PACKET - 1) OR (XRECV > PACKET)            ' if packet number is in error
      MODE := 0                                            '   Set to state zero
      QUIT                                                 '   exit this repeat loop
    ELSE
      PACKET := XRECV                                      ' next Packet Number
      
    XRECV := PC.rx                                         ' 1's complement Packet Number   
    TCNT := 0                                              ' set for start of receive buffer
    REPEAT 128                                             ' Get 128 Bytes from Sender
      XBUF[TCNT] := PC.rx                                  '  get character from sender in put in file transfer buffer
      CRC += XBUF[TCNT++]                                  '  update CRC and incriment file transfer buffer
    
    XRECV := PC.rx                                         ' get CRC from sender
    IF (XRECV <> CRC)                                      ' If CRC fail
      IF (RETRY < 10)                                      '  and retrys are less then 10
        PC.tx(NAK)                                         '    Tell sender we got a bad packet 
        RETRY++                                            '    incriment re-try count 
        REPEAT                                             '    Check for New Packet 
          XRECV := PC.rx                                   '    get character from sender
          IF (XRECV == SOH)                                '    check character for packet start
            QUIT                                           '    exit this repeat loop
            
      ELSE                                                 ' Too many retrys
          MODE := 0                                        '   set tranfer mode to fail
          QUIT                                             '   exit this repeat loop
          
    ELSE                                                   ' Packet looks OK, lets wirte it to the SD Card
      PC.tx(ACK)                                           '   Send ACK to sender
      len := 128
      if ascii > 0
        repeat ii from 0 to len-1
          if byte[@XBUF+ii] == ascii
            len := ii
            byte[@XBUF+len] := 0
            quit                   
      sdfat.pwrite(@XBUF, len)                             '       first packet twice???, flag is used to filter and discard
      PACKET++                                             '       the first xmodem packet.
      
      REPEAT                                               ' Check where we are in the file transfer process
        XRECV := PC.rx
        IF (XRECV == EOT)                                  ' End of transmission ?
          PC.tx(ACK)                                       '   send ACK
          waitcnt(clkfreq/10 + cnt)                        '   wait 100mS
          PC.tx (ACK)                                      '   send ACK again
          MODE := 2                                        '   Set mode to completed successfully
          QUIT                                             '   exit this repeat loop
     
        IF (XRECV == CAN)                                  ' Sender Cancelled transmission?
          MODE := 0                                        '   Transfer aborted or incomplete
          QUIT                                             '   exit this repeat loop

        IF (XRECV == SOH)                                  ' Next packet in the file   
          QUIT

  PC.str(@Newline)                                         ' Send CR & LF to sender
     
  IF (MODE == 2)                                           ' Notify sender of status of file transfer
    PC.str (@Success)                                      '   Upload was a success
  ELSE
    PC.str (@Failure)                                      '   Failure in file transfer

  sdfat.pclose                                             ' Close data file & exit

{  
  if ascii          
    sdfat.popen(@RX_Data, "r")
    PC.str(string($d,"ASCII dump:",$d))
    repeat
      temp := sdfat.pgetc
      if temp < 0
        quit
      PC.tx(temp)      
    PC.tx($d)      

    sdfat.pclose 
}


PRI getString(buf,lim,trm) | ch, ii
  ii := 0
  byte[buf] := 0
  repeat
    ch := PC.rx
    if ch == trm
      quit
    if ii > 0 and ch == BS
      ii--
    else
      byte[buf+ii] := ch
      ii++
    byte[buf+ii] := 0
    PC.tx(ch)
    if ii => lim
      quit
  return buf


      
dat

  ' File name of your Propeller application saved to SD card  
  FileName   byte "BOOTFILE.TXT", 0                       

  ' Program Text Strings:
  SdCardErr  byte "FATAL ERROR --- Can't Mount SD card --- PROGRAM HALTED", CR, LF, 0

  Ready      byte "Ready to receive file via XMODEM.", CR,LF, 0

  Success    byte "--- Transfer completed succesfully.",CR, LF, 0

  Failure    byte "--- Transfer aborted or incomplete.",CR, LF, 0

  GetName    byte "Input File Name: ", 0
  IsAscii    byte " ASCII [y/n] ? ", 0
  
  Newline    byte CR, LF, 0
 